/**
 * abssqr.com Inc. 
 * Copyright (c) 2017-2018 All Rights Reserved. 
 */
package com.abssqr.plat.biz.interceptor;

import com.abssqr.plat.biz.validate.ValidateUtil;
import com.abssqr.plat.biz.validate.ValidatorError;
import com.abssqr.plat.common.facade.base.BaseParam;
import com.abssqr.plat.common.facade.enums.CommonErrorCodeEnum;
import com.abssqr.plat.common.facade.validation.Valid;
import com.abssqr.plat.common.model.exception.AbssqrInvalidParamException;
import com.abssqr.plat.common.util.interceptor.AbstractInterceptor;
import com.abssqr.plat.common.util.loggers.LogInfo;
import com.abssqr.plat.common.util.thread.ThreadLocalCleaner;
import com.general.system.common.model.exception.BaseException;
import com.general.system.common.util.LogUtil;
import org.aopalliance.intercept.MethodInvocation;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Component;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;

/**
 *
 * @author zhenxuan.luo
 * @version com.loan.biz.service.inteceptor: BizServiceInterceptor.java, v 0.1 2018-12-19 AM11:38 zhenxuan.luo Exp $ 
 */
@Component
public class BizServiceInterceptor extends AbstractInterceptor {
    
    /** logger */
    private static final Logger LOGGER = LoggerFactory.getLogger(BizServiceInterceptor.class);
    
    @Override
    public Object businessInvoke(MethodInvocation invocation) throws Throwable {
        Object result = null;
        Throwable ex = null;
        // 初始化调用者信息
        LogInfo logInfo = initDigestLogCallerInfo(invocation);
        
        try {
            // 打印请求日志
            BizLogUtil.logBizReqeust(logInfo, invocation);
            
            // 检查请求参数合法性
            result = checkRequestParam(invocation);
            if (result != null) {
                // 请求参数非法,直接返回非法结果
                ex = new IllegalArgumentException("请求参数非法");
                return result;
            }
            
            // 确保租户id,zoneUid设置到上下文中
            result = invocation.proceed();
        } catch (BaseException e) {
            LogUtil.error(LOGGER, e, "服务调用[{0}]发生业务异常", logInfo);
            ex = e;
            throw e;
            //result = InterceptorUtil.processError(LOGGER, invocation, e.getMessage(), e);
        } catch (Throwable e) {
            LogUtil.error(LOGGER, e, "服务调用[{0}]发生未知异常", logInfo);
            ex = e;
            throw e;
            // result = InterceptorUtil.processError(LOGGER, invocation, e);
        } finally {
            // 防止中间出现异常未release
            
            // 打印业务摘要日志
            BizLogUtil.logBizDigest(logInfo, InterceptorUtil.getEventCode(invocation), ex);
            
            // 清除本地线程变量
            ThreadLocalCleaner.clean();
        }
        return result;
    }
    
    /**
     * 请求参数校验
     *
     * @param invocation
     * @return 校验成功返回null，否则返回失败结果
     */
    private Object checkRequestParam(MethodInvocation invocation)
            throws InstantiationException, IllegalAccessException {
        try {
            
            Annotation[][] parameterAnnotations = invocation.getMethod().getParameterAnnotations();
            for (int i = 0; i < parameterAnnotations.length; i++) {
                for (int j = 0; j < parameterAnnotations[i].length; j++) {
                    if (parameterAnnotations[i][j] instanceof Valid) {
                        Valid param = (Valid) parameterAnnotations[i][j];
                        Object currentArg = invocation.getArguments()[i];
                        //参数为null
                        if (null == currentArg) {
                            if (param.nullable()) {
                                break;
                            } else {
                                String errorMsg = "请求参数不能为null ["
                                                  + invocation.getMethod().getDeclaringClass()
                                                            .getSimpleName() + "."
                                                  + invocation.getMethod().getName() + "]";
                                LogUtil.warn(LOGGER, errorMsg);
                                return buildFailedResult(invocation, true, errorMsg);
                            }
                        }
                        ValidatorError errors = ValidateUtil.validate(currentArg);
                        if (errors.hasErrors()) {
                            String errorMsg = ValidateUtil.collectErrorsLog(currentArg, errors);
                            LogUtil.warn(LOGGER, "非法的请求参数：" + errorMsg);
                            return buildFailedResult(invocation, true, "非法的请求参数" + errorMsg);
                        }
                        if (currentArg instanceof BaseParam) {
                            ((BaseParam) currentArg).checkParam();
                        }
                    }
                }
            }
            return null;
        } catch (AbssqrInvalidParamException e) {
            throw e;
        } catch (IllegalArgumentException e) {
            LogUtil.error(LOGGER, e, "非法的请求参数：" + e.getMessage());
            return buildFailedResult(invocation, true, "非法的请求参数：" + e.getMessage());
        } catch (Throwable e) {
            LogUtil.error(LOGGER, e, "请求参数校验时发生未知错误：" + e.getMessage());
            return buildFailedResult(invocation, false, "请求参数校验时发生未知错误！");
        }
    }
    
    /**
     * 构造参数检查失败时的结果
     *
     * @param invocation
     * @param invalidRequestParam
     * @param errorMsg
     * @return
     */
    private Object buildFailedResult(MethodInvocation invocation, boolean invalidRequestParam,
                                     String errorMsg) throws IllegalAccessException, InstantiationException {
        
        return InterceptorUtil.processError(LOGGER, invocation, errorMsg,
                invalidRequestParam ? CommonErrorCodeEnum.INVALID_PARAM : CommonErrorCodeEnum.SYS_ERROR, null);
        
    }
    
    /**
     * 构建调用方基础信息
     *
     * @param invocation
     */
    private LogInfo initDigestLogCallerInfo(MethodInvocation invocation) {
        
        //设置日志摘要信息
        LogInfo logInfo = new LogInfo();
        
        Method method = invocation.getMethod();
        logInfo.setFacadeName(method.getDeclaringClass().getSimpleName());
        logInfo.setMethodName(method.getName());
        
        return logInfo;
    }
}