/**
 * abssqr.com Inc.
 * Copyright (c) 2017-2019 All Rights Reserved.
 */
package com.abssqr.plat.biz.shared.scheduler.job.plan;

import java.util.Date;
import java.util.List;

import com.abssqr.plat.biz.shared.handler.PlanSettleHandler;
import com.abssqr.plat.biz.shared.scheduler.BaseQuartzJob;
import com.abssqr.plat.common.facade.enums.PlanStatusEnum;
import com.abssqr.plat.common.facade.model.plan.PlanBase;
import com.abssqr.plat.common.model.repo.plan.PlanRepo;
import com.abssqr.plat.core.service.accounting.OrgAcctDayComponent;

import com.general.system.bpm.enums.ApproveStatusEnum;
import com.general.system.common.util.DateTimeUtil;
import com.general.system.common.util.LogUtil;
import com.general.system.common.util.VarChecker;
import com.general.system.tool.util.ToolUtil;
import com.google.common.collect.Lists;
import org.quartz.DisallowConcurrentExecution;
import org.quartz.PersistJobDataAfterExecution;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.support.TransactionCallbackWithoutResult;
import org.springframework.transaction.support.TransactionTemplate;

/**
 * 计划资产需求统计任务
 *
 * @author zhenxuan.luo
 * @version com.abssqr.plat.biz.shared.scheduler.job.plan: PlanPackGenJob.java, v 0.1 2019-06-18 21:19 zhenxuan.luo
 * Exp $
 */
@PersistJobDataAfterExecution
@DisallowConcurrentExecution
@Component
public class PlanDailyJob extends BaseQuartzJob {

    private final static Logger LOGGER = LoggerFactory.getLogger(PlanDailyJob.class);

    @Autowired
    private PlanSettleHandler     planSettleHandler;
    @Autowired
    private TransactionTemplate   transactionTemplate;
    @Autowired
    private PlanRepo              planRepo;
    @Autowired
    private OrgAcctDayComponent   orgAcctDayComponent;
    /**
     * 每天 45 7/20 * * * ?
     */
    @Value("${sync.quartz.plan.daily}")
    private String                cron;

    @Override
    protected String getCronExpr() {
        return cron;
    }

    @Override
    protected boolean isRequestRecovery() {
        return true;
    }

    @Override
    public void doExecute() {
        Date curDate = orgAcctDayComponent.getTrfAcctDay().getStandardDate();
        LogUtil.info(LOGGER, "[DailyJob]-[{0}]计划每日处理任务start", curDate);

        // 查询计划
        List<PlanBase> planBaseList = planRepo.getPlansByStatusList(null,
            Lists.newArrayList(PlanStatusEnum.ONGOING, PlanStatusEnum.PREPARING));

        if (ToolUtil.isEmpty(planBaseList)) {
            return;
        }
        for (PlanBase planEntity : planBaseList) {
            transactionTemplate.execute(new TransactionCallbackWithoutResult() {
                @Override
                protected void doInTransactionWithoutResult(TransactionStatus status) {
                    try {
                        //执行计划每日任务
                        lockEntityAndExecute(curDate, planEntity);

                    } catch (Exception e) {
                        LogUtil.error(LOGGER, e, "{0}[{1}]日切获取锁失败", planEntity.getPlanType().getDesc(), planEntity);
                    }
                }
            });
        }
    }

    /**
     * 计划状态变更
     *
     * @param curDate
     * @param planEntity
     * @return
     */
    public void lockEntityAndExecute(Date curDate, PlanBase planEntity) {
        //顺序不可逆。如果先执行存续期任务可能会漏掉当日变更为存续期的计划
        LogUtil.info(LOGGER, "{0}[{1}]日初计划状态变更任务执行", planEntity.getPlanType().getDesc(), planEntity.getPlanNo());
        // 承接池计划 加锁
        PlanBase planBase = planRepo.lockPlanByNo(planEntity.getPlanNo());
        VarChecker.checkNotNull(planBase, "计划查询加锁为空[{0}]", planBase.getPlanNo());
        if (PlanStatusEnum.PREPARING == planBase.getStatus()
            && DateTimeUtil.calcDayDiff(curDate, planBase.getBeginDate()) >= 0L
            && planBase.getApproveStatus() == ApproveStatusEnum.PASSED) {
            // 变更状态
            planBase.setStatus(PlanStatusEnum.ONGOING);
        }
        // 执行 计划每日业务
        this.onGoingPlansTask(curDate, planBase);
        // 承接池状态与时间变更
        planBase.setLastUpdateDate(curDate);
        planRepo.updatePlan(planBase);
    }

    /**
     * 存续期任务
     *
     * @param curDate
     * @param planEntity
     */
    private void onGoingPlansTask(Date curDate, PlanBase planEntity) {
        if (PlanStatusEnum.ONGOING == planEntity.getStatus()) {
            //存续期任务执行
            if (planEntity.getLastUpdateDate() != null
                && DateTimeUtil.calcDayDiff(curDate, planEntity.getLastUpdateDate()) <= 0) {
                LogUtil.info(LOGGER, "{0}[{1}]最近更新日期为[{2}]，跳过执行",
                    planEntity.getPlanType().getDesc(), planEntity.getPlanNo(),
                    planEntity.getLastUpdateDate());
                return;
            }

            LogUtil.info(LOGGER, "{0}[{1}]日初任务执行", planEntity.getPlanType().getDesc(), planEntity.getPlanNo());
            /**
             * 以下执行顺序不能调换
             */
            // 结息
            planSettleHandler.execute(planEntity, curDate);

        }
    }
}