package com.abssqr.plat.common.util.base;

import com.abssqr.plat.common.util.utils.DayUtil;
import com.general.system.common.util.DateTimeUtil;
import com.general.system.common.util.SystemDateUtil;
import com.google.common.base.Preconditions;
import org.apache.commons.lang3.StringUtils;

import java.util.Calendar;
import java.util.Date;

/**
 * <p>
 * 只表示年月日的日期类
 * </p>
 * <p>
 * <b>注意时区：</b> 如果系统时区不是东八区代码可能是有问题的！尤其是在日期比较的时候！
 * </p>
 *
 * @author zhenxuan.luo
 * @version com.loan.common.model.common: Day.java, v 0.1 2018-12-01 PM20:27 zhenxuan.luo Exp $
 */
public class Day implements Cloneable {
    /**
     * 一个不可能到达的天数,万年前的最后一天
     */
    public static final Day THE_END_DAY = new Day("99991231");

    /**
     * 内部日期表示，只有年月日，格式为yyyyMMdd
     */
    private final String dayString;

    /**
     * 内部日期表示，只有年月日，格式为yyyy-MM-dd
     */
    private final String daySortString;

    /**
     * 内部日期时间表示，格式为yyyyMMddHHmmss
     */
    private final String dayTimeString;

    /**
     * 内部日期对象
     */
    private final Date date;

    /**
     * 默认构造函数
     * 以当前系统时间构造一个日期对象
     */
    public Day() {
        this(new Date());
    }

    /**
     * 可为null的拷贝
     *
     * @param obj
     * @return
     */
    public static Day cloneNotNull(Day obj) {
        if (obj == null) {
            return null;
        } else {
            return new Day(obj);
        }
    }

    /**
     * 日期字符串格式为yyyyMMdd
     *
     * @param dayString 日期字符串
     */
    public Day(String dayString) {
        Preconditions.checkArgument(!StringUtils.isBlank(dayString) && dayString.length() == 8,
            "日期格式错误：" + dayString);

        this.date = DateTimeUtil.parseFromYMD(dayString);
        if (this.date == null) {
            throw new IllegalArgumentException("日期格式错误：" + dayString);
        }

        this.dayString = dayString;
        this.daySortString = DateTimeUtil.formatY_M_D(this.date);
        this.dayTimeString = dayString + "000000";
    }

    /**
     * 根据给定日期的年月日构造实例，时分秒值被直接忽略
     *
     * @param date 日期
     */
    public Day(Date date) {
        Preconditions.checkArgument(date != null);
        Calendar calendar = Calendar.getInstance();
        calendar.setTime(date);
        if (calendar.get(Calendar.HOUR_OF_DAY) + calendar.get(Calendar.MINUTE) + calendar.get(Calendar.SECOND)
            + calendar.get(Calendar.MILLISECOND) == 0) {
            Date time = new Date();
            Calendar timeCalendar = Calendar.getInstance();
            timeCalendar.setTime(time);
            calendar.set(Calendar.HOUR_OF_DAY, timeCalendar.get(Calendar.HOUR_OF_DAY));
            calendar.set(Calendar.MINUTE, timeCalendar.get(Calendar.MINUTE));
            calendar.set(Calendar.SECOND, timeCalendar.get(Calendar.SECOND));
            calendar.set(Calendar.MILLISECOND, timeCalendar.get(Calendar.MILLISECOND));
        }
        date = calendar.getTime();
        this.dayString = DateTimeUtil.formatYMD(date);
        this.daySortString = DateTimeUtil.formatY_M_D(date);
        this.dayTimeString = DateTimeUtil.formatYMDHMSShort(date);
        this.date = (Date)date.clone();
    }

    /**
     * 根据给定的日期，构造一个同样时间的日期
     *
     * @param day 日期
     */
    public Day(Day day) {
        Preconditions.checkArgument(day != null);

        this.dayString = day.dayString;
        this.dayTimeString = day.dayTimeString;
        this.daySortString = day.daySortString;
        this.date = (Date)day.date.clone();
    }

    /**
     * Getter method for property <tt>date</tt>.
     *
     * @return property value of date
     */
    public Date getDate() {
        return date;
    }

    /**
     * Getter method for property <tt>date</tt>.
     *
     * @return property value of date
     */
    public Date getStandardDate() {
        return DateTimeUtil.getStandardDate(date);
    }

    /**
     * Getter method for property <tt>date</tt>.
     *
     * @return property value of date
     */
    public String getDateString() {
        return DateTimeUtil.formatY_M_D(date);
    }

    /**
     * 转化为标准日期类
     *
     * @return 日期类
     */
    public Date toDate() {
        return (Date)this.date.clone();
    }

    /**
     * 返回yyyyMMdd格式的日期字符串
     *
     * @return yyyyMMdd格式的日期字符串
     */
    public String getDayString() {
        return dayString;
    }

    /**
     * Getter method for property <tt>dayTimeString</tt>.
     *
     * @return property value of dayTimeString
     */
    public String getDayTimeString() {
        return dayTimeString;
    }

    /**
     * 根据当前日期，获取昨天
     *
     * @return 当前日期的昨天
     */
    public Day getYesterday() {
        return DayUtil.addDays(this, -1);
    }

    /**
     * 根据当前日期，获取明天
     *
     * @return 当前日期的明天
     */
    public Day getTomorrow() {
        return DayUtil.addDays(this, 1);
    }

    /**
     * 返回yyyyMMdd格式的日期字符串
     *
     * @return yyyyMMdd格式的日期字符串
     */
    @Override
    public String toString() {
        return this.dayString;
    }

    /**
     * 判断两个日期是否表示同一天
     *
     * @param o 对象
     * @return true如果表示同一天，否则false
     */
    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if (!(o instanceof Day)) {
            return false;
        }
        Day day = (Day)o;
        return dayString.equals(day.dayString);
    }

    /**
     * 判断两个日期是否表示同一天
     *
     * @param day 待比较的日期
     * @return true如果表示同一天，否则false
     */
    public boolean equals(Day day) {
        return day != null && dayString.equals(day.dayString);
    }

    /**
     * 返回hash码
     *
     * @return hash码
     */
    @Override
    public int hashCode() {
        return dayString.hashCode();
    }

    /**
     * 构造一个表示同一天的类对象
     *
     * @return 表示同一天的类对象
     */
    @Override
    public Day clone() {
        return new Day(this);
    }

    public String getDaySortString() {
        return daySortString;
    }

    public boolean equalSysDate(){
        return DateTimeUtil.isSameDay(date, SystemDateUtil.getSystemDate());
    }


}
