/**
 * abssqr.com Inc.
 * Copyright (c) 2017-2018 All Rights Reserved.
 */

package com.abssqr.plat.core.service.ast;

import com.abssqr.plat.common.dal.mysql.auto.dataobject.AstPackDO;
import com.abssqr.plat.common.facade.enums.AstPackStatusEnum;
import com.abssqr.plat.common.facade.enums.AstPackTradeTypeEnum;
import com.abssqr.plat.common.facade.model.plan.PlanTransferRule;
import com.abssqr.plat.common.facade.model.prod.LoanProduct;
import com.abssqr.plat.common.facade.model.rule.AstScreenRule;
import com.abssqr.plat.common.model.convertor.PacKConvert;
import com.abssqr.plat.common.model.domain.ast.Pack;
import com.abssqr.plat.common.model.repo.ast.PackRepository;
import com.abssqr.plat.common.model.repo.plan.PlanTransferRuleRepo;
import com.abssqr.plat.common.model.repo.prod.LoanProdRepository;
import com.general.system.common.util.LogUtil;
import com.general.system.common.util.VarChecker;
import com.general.system.tool.util.ToolUtil;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import org.apache.commons.collections.CollectionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.transaction.support.TransactionTemplate;

import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * 资产包日切服务
 *
 * @author guozhenhua
 * @version AstPackManager.java, v 0.1 2018-05-16 下午4:43 guozhenhua Exp $
 */
@Component
public class AstPackManager {

    private final static Logger logger = LoggerFactory.getLogger(AstPackManager.class);

    @Autowired
    private TransactionTemplate  transactionTemplate;
    @Autowired
    private LoanProdRepository   loanProdRepository;
    @Autowired
    private PlanTransferRuleRepo planTransferRuleRepo;
    @Autowired
    private PackRepository       packRepository;
    @Autowired
    private LoanProdRepository   prodRepository;

    /**
     * 更新资产包状态
     *
     * @param packs
     * @param packStatus
     * @param oldPackStatus
     */
    public void updatePackStatus(List<Pack> packs, AstPackStatusEnum packStatus, AstPackStatusEnum oldPackStatus) {
        packRepository.updatePackStatus(packs, packStatus, oldPackStatus);
    }

    /**
     * 获取待筛选的包
     *
     * @param orgCodes
     * @param scenDate
     * @param packStatus
     * @return
     */
    public Map<String, List<Pack>> queryPacksByOrgCodes(List<String> orgCodes, Date scenDate,
                                                        AstPackStatusEnum packStatus) {
        // 查询资产包
        List<AstPackDO> astPackDOS = packRepository.selectPackByStatus(scenDate, packStatus, AstPackTradeTypeEnum.AST_TRF_BIZ);
        Map<String, List<Pack>> result = Maps.newHashMap();
        if (CollectionUtils.isNotEmpty(astPackDOS)) {
            for (AstPackDO astPackDO : astPackDOS) {
                Pack pack = PacKConvert.do2Domain(astPackDO);
                PlanTransferRule transferRule = planTransferRuleRepo.getByPlanNo(astPackDO.getPlanNo());
                VarChecker.checkNotNull(transferRule);
                pack.setTransferRule(transferRule);

                List<String> prodCodes = transferRule.getAstScreenRuleList().stream().map(
                        AstScreenRule::getAbssqrProdCode).collect(
                        Collectors.toList());

                for (String prodCode : prodCodes) {
                    LoanProduct product = loanProdRepository.getBybssqrProdCode(prodCode);
                    VarChecker.checkNotNull(product, "产品编码【{0}】不存在", prodCode);
                    String orgCode = product.getManagerOrgCode();
                    if (orgCodes.contains(product.getManagerOrgCode())) {
                        List<Pack> packs = result.get(orgCode);
                        if (packs == null) {
                            packs = Lists.newArrayList();
                            result.put(orgCode, packs);
                        }
                        packs.add(pack);
                    }
                }
            }
        } else {
            LogUtil.info(logger, "没有找到[{0}]日期的待筛选资产包", scenDate);
        }
        return result;
    }

    /**
     * 根据计划编号锁查询资产包
     *
     * @param packCode
     * @return
     */
    public Pack lockByPackCode(String packCode) {
        AstPackDO astPackDO = packRepository.lockByPackCode(packCode);
        return convert2Domain(astPackDO);
    }


    /**
     * 筛选计划最近一次的资产包
     *
     * @param planNo
     * @param scenDate
     * @return
     */
    public Pack getLastDayPackByPlanNo(String planNo, Date scenDate) {
        AstPackDO lastDayPackByPlanNo = packRepository.getLastDayPackByPlanNo(planNo, scenDate, AstPackTradeTypeEnum.AST_TRF_BIZ);
        return convert2Domain(lastDayPackByPlanNo);
    }

    /**
     * 根据管理机构获取资产包
     *
     * @param managerOrgCode
     * @param accountDate
     * @param status
     * @return
     */
    public List<Pack> getByManagerOrgCode(String managerOrgCode, Date accountDate, AstPackStatusEnum status) {
        List<AstPackDO> astPackDOList = packRepository.getByManagerOrgCode(managerOrgCode, accountDate, status, AstPackTradeTypeEnum.AST_TRF_BIZ);
        return convert2DomainList(astPackDOList);
    }

    /**
     * 根据packCode获取数据
     *
     * @param packCode
     * @return
     */
    public Pack getByPackNo(String packCode) {
        AstPackDO packDO = packRepository.getByPackNo(packCode);
        return convert2Domain(packDO);
    }


    private List<Pack> convert2DomainList(List<AstPackDO> astPackDOList) {
        if (ToolUtil.isEmpty(astPackDOList)){
            return Lists.newArrayList();
        }
        List<Pack> packList = PacKConvert.convert2DoList(astPackDOList);
        packList.forEach(item -> {
            fillPack(item);
        });
        return PacKConvert.convert2DoList(astPackDOList);

    }

    private Pack convert2Domain(AstPackDO astPackDO) {
        Pack pack = PacKConvert.do2Domain(astPackDO);
        fillPack(pack);
        return pack;
    }


    /**
     * 填充数据
     *
     * @param pack
     */
    private void fillPack(Pack pack) {
        if (pack == null){
            return;
        }
        PlanTransferRule transferRule = planTransferRuleRepo.getByPlanNo(pack.getPlanNo());
        VarChecker.checkNotNull(transferRule);
        pack.setTransferRule(transferRule);
    }


}
