/**
 * abssqr.com Inc.
 * Copyright (c) 2017-2020 All Rights Reserved.
 */

package com.abssqr.plat.core.service.ast;

import com.abssqr.plat.common.dal.mysql.auto.dataobject.AstPackDO;
import com.abssqr.plat.common.facade.enums.CommonErrorCodeEnum;
import com.abssqr.plat.common.facade.enums.EventCodeEnum;
import com.abssqr.plat.common.facade.enums.OwnStatusEnum;
import com.abssqr.plat.common.facade.enums.PlanTypeEnum;
import com.abssqr.plat.common.facade.enums.ProfitTypeEnum;
import com.abssqr.plat.common.facade.enums.TransCodeEnum;
import com.abssqr.plat.common.facade.model.plan.PlanBase;
import com.abssqr.plat.common.facade.model.plan.PlanTransferRule;
import com.abssqr.plat.common.facade.model.prod.LoanProduct;
import com.abssqr.plat.common.facade.model.rule.AstPriceRule;
import com.abssqr.plat.common.model.convertor.LoanAstBuyBackPackConvert;
import com.abssqr.plat.common.model.domain.ast.Loan;
import com.abssqr.plat.common.model.domain.ast.LoanAst;
import com.abssqr.plat.common.model.domain.ast.LoanAstBuyBackPack;
import com.abssqr.plat.common.model.domain.ast.LoanAstGroup;
import com.abssqr.plat.common.model.domain.ast.LoanAstProfit;
import com.abssqr.plat.common.model.domain.ast.LoanAstProfitGroup;
import com.abssqr.plat.common.model.domain.ast.LoanBak;
import com.abssqr.plat.common.model.domain.ast.LoanInstalDomain;
import com.abssqr.plat.common.model.domain.clearing.PlanClearingLogKey;
import com.abssqr.plat.common.model.domain.loanTrans.LoanAstLog;
import com.abssqr.plat.common.model.domain.loanTrans.LoanAstProfitTransDetail;
import com.abssqr.plat.common.model.domain.loanTrans.LoanAstTransDetail;
import com.abssqr.plat.common.model.domain.loanTrans.LoanInstalTransDetail;
import com.abssqr.plat.common.model.domain.loanTrans.LoanTransDetail;
import com.abssqr.plat.common.model.domain.loanTrans.TransInfoTemplate;
import com.abssqr.plat.common.model.domain.rule.LoanAstBalAndPrice;
import com.abssqr.plat.common.model.domain.rule.LoanAstPriceCalcParam;
import com.abssqr.plat.common.model.domain.rule.LoanAstPriceResult;
import com.abssqr.plat.common.model.domain.rule.LoanAstProfitBalAndPrice;
import com.abssqr.plat.common.model.domain.task.TrTask;
import com.abssqr.plat.common.model.domain.trade.AstTradeParam;
import com.abssqr.plat.common.model.exception.AbssqrBizException;
import com.abssqr.plat.common.model.repo.ast.LoanRepository;
import com.abssqr.plat.common.model.repo.ast.PackRepository;
import com.abssqr.plat.common.model.repo.plan.PlanRepo;
import com.abssqr.plat.common.model.repo.plan.PlanTransferRuleRepo;
import com.abssqr.plat.common.model.repo.prod.LoanProdRepository;
import com.abssqr.plat.common.model.repo.rule.AstPriceRuleRepo;
import com.abssqr.plat.common.model.utils.LoanAstUtil;
import com.abssqr.plat.common.util.profiler.PerformTrace;
import com.abssqr.plat.core.service.accounting.OrgAcctDayComponent;
import com.abssqr.plat.core.service.plan.PlanClearingManager;
import com.abssqr.plat.core.service.rule.AstPriceRuleComponent;
import com.abssqr.plat.core.service.rule.AstScreenRuleComponent;
import com.abssqr.plat.core.service.system.TimeStampUtil;
import com.abssqr.plat.core.service.util.AstPriceCalcAmtUtil;
import com.general.system.common.model.Money;
import com.general.system.common.util.LogUtil;
import com.general.system.common.util.MessageUtil;
import com.general.system.common.util.VarChecker;
import com.google.common.collect.Maps;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.List;
import java.util.Map;

/**
 * 回购业务
 *
 * @author xiachenxiang
 * @version com.abssqr.plat.core.service.ast: LoanAstBuyBackManager.java, v 0.1 2020-02-11 5:00 PM xiachenxiang Exp $
 */
@Component
public class LoanAstBuyBackManager {

    private final static Logger logger = LoggerFactory.getLogger(LoanAstBuyBackManager.class);

    @Autowired
    private AstPoolManager         astPoolManager;
    @Autowired
    private PackRepository         packRepository;
    @Autowired
    private LoanProdRepository     prodRepository;
    @Autowired
    private AstPriceRuleRepo       astPriceRuleRepo;
    @Autowired
    private PlanClearingManager    planClearingManager;
    @Autowired
    private PlanRepo               planRepo;
    @Autowired
    private LoanAstAssembler       loanAstAssembler;
    @Autowired
    private LoanRepository         loanRepository;
    @Autowired
    private AstScreenRuleComponent astScreenRuleComponent;
    @Autowired
    private AstPriceRuleComponent  astPriceRuleComponent;
    @Autowired
    private PlanTransferRuleRepo   planTransferRuleRepo;
    @Autowired
    private OrgAcctDayComponent    orgAcctDayComponent;

    /**
     * 根据packCode获取数据
     *
     * @param packCode
     * @return
     */
    @PerformTrace
    public LoanAstBuyBackPack getByPackCode(String packCode) {
        AstPackDO astPackDO = packRepository.getByPackNo(packCode);
        LoanAstBuyBackPack loanAstBuyBackPack = LoanAstBuyBackPackConvert.do2Domain(astPackDO);
        fillPack(loanAstBuyBackPack);
        return loanAstBuyBackPack;
    }

    /**
     * 填充数据
     *
     * @param pack
     */
    private void fillPack(LoanAstBuyBackPack pack) {
        // 1.产品
        LoanProduct loanProduct = prodRepository.getBybssqrProdCode(pack.getLoanProduct().getAbssqrProdCode());
        VarChecker.checkNotNull(loanProduct, "[{0}]产品不存在", getClass().getSimpleName(), pack.getLoanProduct().getAbssqrProdCode());
        pack.setLoanProduct(loanProduct);
        // 2.定价规则
        List<AstPriceRule> priceRuleList = astPriceRuleRepo.getByAbsProdCode(pack.getLoanProduct().getAbssqrProdCode());
        VarChecker.checkNotEmpty(priceRuleList, "[{0}]产品定价规则为空", pack.getLoanProduct().getAbssqrProdCode());
        pack.setAstPriceRuleList(priceRuleList);
        // 3.转让规则
        PlanTransferRule transferRule = planTransferRuleRepo.getByPlanNo(pack.getPlanNo());
        VarChecker.checkNotNull(transferRule, "[{0}]计划转让规则为空", pack.getLoanProduct().getAbssqrProdCode());
        pack.setTransferRule(transferRule);
    }

    /**
     * @param trTask
     * @param tradeParam
     * @return
     */
    @PerformTrace
    public boolean astBuyBackTrade(TrTask trTask, AstTradeParam tradeParam) {
        // 获取资产
        Loan loan = tradeParam.getLoan();
        // 资产包
        LoanAstBuyBackPack pack = tradeParam.getPack();
        // 2 转让规则
        PlanTransferRule transferRule = pack.getTransferRule();

        // 获取资产转入机构计划
        PlanBase inPlan = planRepo.getPlanByNo(pack.getPlanNo());

        // 筛选所有期数机构资产
        LoanAstGroup loanAstGroup = loan.filterLoanAstByOrg(trTask.getTfrOut(), trTask.getTermNo());
        // 3 校验回购筛选规则 不匹配直接跳出
        boolean isValid = astScreenRuleComponent.checkBuyback(loan, loanAstGroup, trTask, pack);
        if (!isValid) {
            return isValid;
        }
        // 保证一个时间戳，方便汇总笔数
        String timeStamp = TimeStampUtil.getDefaultTimeStamp();

        // 4 处理资产后沟卖出 构造交易详情
        TransInfoTemplate transInTemplate = new TransInfoTemplate();
        transInTemplate.setTransTime(orgAcctDayComponent.getTrfAcctDay().getStandardDate());
        transInTemplate.setAccountDate(trTask.getTfrDt());
        transInTemplate.setEventCode(EventCodeEnum.BACK_OUT);
        transInTemplate.setOutBizNo(trTask.getBatchNo());
        transInTemplate.setBizNo(trTask.getSeqNo());
        transInTemplate.setTimestamp(timeStamp);
        LoanTransDetail loanTransDetail = LoanTransDetail.createLoanAstTxnDtl(loan, transInTemplate);

        // 5 获取待转分期
        Map<Integer, LoanInstalTransDetail> instalTransDetailMap = Maps.newHashMap();
        boolean isTransfer = false;
        LoanAstBalAndPrice checkAmt = new LoanAstBalAndPrice();
        for (LoanAst loanAst : loanAstGroup.getLoanAsts()) {
            // 获取机构资产对应分期
            LoanInstalDomain refInstal = loanAst.getRefInstal();
            // 如果不是逾期或正常的分期，不做转让（对应过滤的状态是：结清，冲正，取消）
            if (!refInstal.getStatus().isActive()) {
                continue;
            }
            // 当前机构 != 转出机构，理论上不会发生，除非过滤处有bug
            if (!StringUtils.equals(loanAst.getOwnOrg(), trTask.getTfrOut())) {
                throw new AbssqrBizException(
                        MessageUtil.formatMsg("资产回购交易[{0}]过程中发现转出机构不符[{1}]，停止执行",
                                trTask, loanAst), CommonErrorCodeEnum.INVALID_INTERNAL_MODEL, false);
            }

            isTransfer = true;
            // 处理分期
            LoanInstalTransDetail instalTransDetail = instalTransDetailMap.get(refInstal.getTermNo());
            if (null == instalTransDetail) {
                // 要交易的分期构建交易明细
                instalTransDetail = loanTransDetail.addLoanInstalTransDtl(refInstal);
                instalTransDetailMap.put(refInstal.getTermNo(), instalTransDetail);
            }

            // 6 组装定价规则计算当前机构资产计价
            LoanAstPriceCalcParam param = new LoanAstPriceCalcParam();
            param.setTransIn(pack.getPlanNo());
            param.setTransOut(trTask.getTfrOut());
            param.setLoanAst(loanAst);
            param.addPriceRuleByTask(trTask, pack.getLoanProduct().getAbssqrProdCode());
            param.setAstTradeType(tradeParam.getAstTradeType());
            // 资产五级分类
            param.setCondition(loan.getLoanRisk().getRiskClass().getCode());
            LoanAstPriceResult result = astPriceRuleComponent.execute(param);
            // 为校验准备
            checkAmt.merge(result);

            /**
             * 7 执行转入转出
             * 7.1 处理资产本金转入
             */
            // 回购资产创建新的机构资产
            LoanAst loanAstIn = LoanAstUtil.buildForTransfer(refInstal, loanAst, pack, trTask.getTfrDt()
                    , null, null, transferRule.getTransferNo());
            loanAstIn.setInDate(trTask.getTfrDt());
            LoanAstTransDetail astTransDetailIn = LoanAstTransDetail.createTransDtl(
                    loanAstIn, instalTransDetail, EventCodeEnum.BACK_IN);
            // 机构资产转入处理本金，发生额为原本金，成本为计价金额
            astTransDetailIn.transIn(result.getBal().getAllPrin(), result.getPrice().getSum());
            // 转入机构资产记录扩展信息
            loanAstIn.setExtInf(result.getExtText());

            /**
             * 7.2 处理资产本金转出
             */
            LoanAstTransDetail astTransDetailOut = LoanAstTransDetail.createTransDtl(
                    loanAst, instalTransDetail, EventCodeEnum.BACK_OUT);
            // 机构资产转出处理本金
            astTransDetailOut.transferOut();
            // 转出机构资产记录扩展信息
            loanAst.setExtInf(result.getExtText());
            // 7.3 分期交易处理转让本金
            instalTransDetail.instalBuyback(result.getBal());
            /**
             * 7.4 处理资产收益的转入、转出
             */
            // 根据机构资产当期所有收益进行定价计算
            for (LoanAstProfitGroup astProfitGroup : loanAst.getLoanAstProfits().values()) {
                // 获取持有收益
                LoanAstProfit astProfit = astProfitGroup.getOwnAstProfit();
                // 取定价结果
                LoanAstProfitBalAndPrice profitBalAndPrice = result.getProfitBalAndPrice(astProfit);

                // 7.4.1 处理转入机构资产收益 对于新创建的机构资产，初始化机构收益
                LoanAstProfit inAstProfit = loanAstIn.getLoanAstProfit(astProfit.getSrcTermNo(), ProfitTypeEnum.OWN);
                if (inAstProfit == null) {
                    // 根据原分期 创建对应收益
                    inAstProfit = LoanAstUtil.buildDefaultProfit(
                            loan.getLoanInstalMap().get(astProfit.getSrcTermNo()), loanAstIn, loanAstIn.getInDate());
                }
                LoanAstProfitTransDetail profitTransDetailIn =
                        LoanAstProfitTransDetail.createTransDtl(inAstProfit, instalTransDetail, EventCodeEnum.BACK_IN);
                profitTransDetailIn.transIn(astProfit, profitBalAndPrice);

                // 7.4.2 处理转出机构资产收益
                LoanAstProfitTransDetail profitTransDetailOut =
                        LoanAstProfitTransDetail.createTransDtl(astProfit, instalTransDetail, EventCodeEnum.BACK_OUT);
                profitTransDetailOut.transOut(profitBalAndPrice);
            }
        }
        if (!isTransfer) {
            LogUtil.info(logger, "资产回购交易没有可回购资产astNo[{0}]out[{1}]=>in[{2}]",
                    trTask.getAstNo(), trTask.getTfrOut(), inPlan.getPlanNo());
            return false;
        }

        // 8 校验当前机构资产价格与计价金额 大于直接跳出
        if (AstPriceCalcAmtUtil.checkGreaterThan(checkAmt, trTask)) {
            LogUtil.info(logger, "资产回购交易发生额大于计价金额amt[{0}]trTask[{1}]",
                    checkAmt, trTask);
            return false;
        }

        // 9 登记回购清算流水
        planClearingManager.processClearingForLoanAst(
                loanTransDetail.getLoanInstalTransDetails(), TransCodeEnum.BUY_BACK_OUT, transInTemplate.getAccountDate(),
                instalTransDetail -> {
                    Map<PlanClearingLogKey, Money> ret = Maps.newHashMap();
                    // 处理机构资产 上对应的 清算金额ClrgAmt 包含了利息 ，机构资产ClrgAmt = 回购计价金额（本利罚）
                    for (LoanAstTransDetail astTransDetail : instalTransDetail.getLoanAstTransDetails()) {
                        LoanAstLog astLog = astTransDetail.getActiveLog();
                        if (EventCodeEnum.BACK_OUT == astLog.getEventCode()) {
                            continue;
                        }
                        PlanClearingLogKey key = new PlanClearingLogKey(trTask.getTfrOut(), astLog.getOwnOrg(),
                                TransCodeEnum.BUY_BACK_OUT);
                        if (!ret.containsKey(key)) {
                            ret.put(key, new Money());
                        }
                        // 本金
                        ret.get(key).addTo(astLog.getClrgAmt());
                    }
                    return ret;
                });

        // 10 回购交易后资产归属状态
        if (inPlan.getPlanType() == PlanTypeEnum.SPV) {
            boolean isAllTBack = true;
            for (LoanInstalDomain loanInstal : loan.getLoanInstalList()) {
                // 如果不是逾期或正常的分期，不做转让（对应过滤的状态是：结清，冲正，取消）
                if (!loanInstal.getStatus().isActive()) {
                    continue;
                }
                // 判断是否全部归属放款池资产
                LoanAstGroup instalAstGroup = loanInstal.filterLoanAst(
                        loanInstal.getAstOrg().getAstOwnership().getLenderOrgCode());
                if (!instalAstGroup.isAllAst()) {
                    isAllTBack = false;
                    break;
                }
            }
            if (isAllTBack) {
                // 全部回购 将资产归属状态改为初始状态
                loan.getAstOrg().getAstOwnership().setOwnStatus(OwnStatusEnum.INIT);
            } else {
                // 放款池部分回购后 将资产归属状态改为部分转出
                loan.getAstOrg().getAstOwnership().setOwnStatus(OwnStatusEnum.PARTIAL_TRANS_OUT);
            }
        }
        // 11 更新资产
        loanTransDetail.processBeforeUpdate();
        loanAstAssembler.store(loan, trTask.getTfrDt(), false);
        // 对于转入机构的第一次备份
        List<LoanBak> loanBaks = LoanTransDetail.buildBaks(loan, transInTemplate.getAccountDate());
        loanRepository.backup(loanBaks);

        // 12 放款池回购时 创建资产池资产
        if (inPlan.getPlanType() == PlanTypeEnum.SPV) {
            astPoolManager.buildAstPoolByBuyBack(loan, transInTemplate.getAccountDate());
        }
        LogUtil.info(logger, "[PlanBuyBackManager]回购交易完成astNo[{0}]out[{1}]=>in[{2}]",
                trTask.getAstNo(), trTask.getTfrOut(), inPlan.getPlanNo());
        return true;
    }
}
