/**
 * abssqr.com Inc.
 * Copyright (c) 2017-2017 All Rights Reserved.
 */

package com.abssqr.plat.core.service.ast;

import com.abssqr.plat.common.facade.enums.*;
import com.abssqr.plat.common.facade.model.plan.Plan;
import com.abssqr.plat.common.facade.model.plan.PlanTransferRule;
import com.abssqr.plat.common.facade.model.plan.PlanAstPriceRule;
import com.abssqr.plat.common.model.domain.ast.Loan;
import com.abssqr.plat.common.model.domain.ast.LoanAst;
import com.abssqr.plat.common.model.domain.ast.LoanAstGroup;
import com.abssqr.plat.common.model.domain.ast.LoanBak;
import com.abssqr.plat.common.model.domain.ast.LoanInstalDomain;
import com.abssqr.plat.common.model.domain.ast.Pack;
import com.abssqr.plat.common.model.domain.clearing.PlanClearingLogKey;
import com.abssqr.plat.common.model.domain.loanTrans.LoanAstLog;
import com.abssqr.plat.common.model.domain.loanTrans.LoanAstProfitTransDetail;
import com.abssqr.plat.common.model.domain.loanTrans.LoanAstTransDetail;
import com.abssqr.plat.common.model.domain.loanTrans.LoanInstalTransDetail;
import com.abssqr.plat.common.model.domain.loanTrans.LoanTransDetail;
import com.abssqr.plat.common.model.domain.loanTrans.TransInfoTemplate;
import com.abssqr.plat.common.model.domain.task.TrTask;
import com.abssqr.plat.common.model.exception.AbssqrBizException;
import com.abssqr.plat.common.model.repo.ast.LoanRepository;
import com.abssqr.plat.common.model.utils.AstPriceUtil;
import com.abssqr.plat.common.model.utils.LoanAstUtil;
import com.abssqr.plat.common.util.profiler.PerformTrace;
import com.abssqr.plat.core.service.accounting.OrgAcctDayComponent;
import com.abssqr.plat.core.service.plan.PlanClearingManager;
import com.abssqr.plat.core.service.plan.PlanManager;
import com.abssqr.plat.core.service.rule.AstScreenRuleComponent;
import com.abssqr.plat.core.service.system.TimeStampUtil;
import com.general.system.common.model.Money;
import com.general.system.common.util.MessageUtil;
import com.general.system.common.util.SystemDateUtil;
import com.google.common.collect.Maps;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.math.BigDecimal;
import java.util.List;
import java.util.Map;

/**
 * 贷款资产转让管理器
 *
 * @author guozhenhua
 * @version LoanAstTrManager.java, v 0.1 2017-12-18 下午2:37 guozhenhua Exp $
 */
@Component
public class LoanAstTrManager {

    private final static Logger LOGGER = LoggerFactory.getLogger(LoanAstTrManager.class);

    @Autowired
    private LoanRepository         loanRepository;
    @Autowired
    private LoanAstAssembler       loanAstAssembler;
    @Autowired
    private AstPackManager         astPackManager;
    @Autowired
    private PlanClearingManager    planClearingManager;
    @Autowired
    private AstScreenRuleComponent astScreenRuleComponent;
    @Autowired
    private PlanManager            planManager;
    @Autowired
    private OrgAcctDayComponent    orgAcctDayComponent;

    /**
     * 执行信贷资产转让，保证外层已经起事务
     *
     * @param trTask
     * @return 如果返回false，表示该笔资产不能转让给指定计划，转让失败。如果返回true，表示转让成功。抛异常可重试
     */
    @PerformTrace
    public boolean transfer(TrTask trTask) {

        LOGGER.info("[LoanAstTrManager]执行转让任务：" + trTask);

        /**
         * 1. 准备转让信息：资产和资产包
         */
        // 获取资产
        Loan loan = loanAstAssembler.lockByAstNo(trTask.getAstNo(), trTask.getTfrDt());
        if (loan == null) {
            LOGGER.warn(MessageUtil.formatMsg("[AstTransferIgnore][{0}]捞不到待转借据", trTask));
            return false;
        }
        // 资产包
        Pack pack = astPackManager.getByPackNo(trTask.getTfrPack());
        // 筛选机构资产
        LoanAstGroup loanAstGroup = loan.filterLoanAstByOrg(trTask.getTfrOut(), trTask.getTermNo());
        // 转让规则
        Plan inPlan = planManager.getPlanFullByNo(pack.getPlanNo());
        PlanTransferRule planTransferRule = inPlan.getPlanTransferRule();
        List<PlanAstPriceRule> astPriceRules = inPlan.getPlanAstPriceRuleList();

        /**
         * 2. 转让规则校验
         */
        boolean isValid = astScreenRuleComponent.checkTransfer(
                loan, loanAstGroup, trTask, pack, planTransferRule);
        if (!isValid) {
            return isValid;
        }

        /**
         * 3. 执行转让
         */
        // 保证一个时间戳，方便汇总笔数
        String timeStamp = TimeStampUtil.getDefaultTimeStamp();

        // 这期简单执行，只做打标
        // 处理转入
        TransInfoTemplate transInTemplate = new TransInfoTemplate();
        transInTemplate.setTransTime(SystemDateUtil.getSystemDate());
        transInTemplate.setAccountDate(trTask.getTfrDt());
        transInTemplate.setEventCode(EventCodeEnum.TRANS_OUT);
        transInTemplate.setOutBizNo(trTask.getBatchNo());
        transInTemplate.setBizNo(trTask.getSeqNo());
        transInTemplate.setTimestamp(timeStamp);
        // 构造交易详情
        LoanTransDetail loanTransDetail = LoanTransDetail.createLoanAstTxnDtl(loan, transInTemplate);

        // 获取待转分期
        Map<Integer, LoanInstalTransDetail> instalTransDetailMap = Maps.newHashMap();

        boolean isTransfer = false;
        for (LoanAst loanAst : loanAstGroup.getLoanAsts()) {
            // 获取机构资产对应分期
            LoanInstalDomain refInstal = loanAst.getRefInstal();
            // 如果不是逾期或正常的分期，不做转让（对应过滤的状态是：结清，冲正，取消）
            if (!refInstal.getStatus().isActive()) {
                continue;
            }

            // 当前机构 != 转出机构，理论上不会发生，除非过滤处有bug
            if (!StringUtils.equals(loanAst.getOwnOrg(), trTask.getTfrOut())) {
                throw new AbssqrBizException(
                        MessageUtil.formatMsg("[AstTransferError]转让[{0}]过程中发现转出机构不符[{1}]，停止执行", trTask, loanAst),
                        CommonErrorCodeEnum.INVALID_INTERNAL_MODEL, false);
            }

            isTransfer = true;
            // 处理分期
            LoanInstalTransDetail instalTransDetail = instalTransDetailMap.get(refInstal.getTermNo());
            if (null == instalTransDetail) {
                instalTransDetail = loanTransDetail.addLoanInstalTransDtl(refInstal);
                instalTransDetailMap.put(refInstal.getTermNo(), instalTransDetail);
            }

            // 定价
            Money cost = loanAst.getPrinBal();
            Money transPrinAmt = loanAst.getPrinBal().clone();
            PriceTypeEnum priceType = inPlan.getPriceType();
            BigDecimal price = BigDecimal.ZERO;

            if (priceType == PriceTypeEnum.SPLIT_PROFIT) {
                PlanAstPriceRule priceRule = AstPriceUtil.filterPrice(astPriceRules, Long.valueOf(loanAst.getTermNo()));
                if (null != priceRule) {
                    price = priceRule.getIntrDiscountRate();
                }
            }

            // 处理转入
            // 如果机构资产已存在（如通过分期调整产生），则直接追加，否则创建新的机构资产
            LoanAst loanAstIn = refInstal.filterLoanAst(pack.getPlanNo(), inPlan.getPlanTransferRule().getTransferNo());
            if (null == loanAstIn) {
                loanAstIn = LoanAstUtil.buildForTransfer(refInstal, loanAst, pack, trTask.getTfrDt(),
                        price, priceType, planTransferRule.getTransferNo());
                loanAstIn.setInDate(trTask.getTfrDt());
            }
            LoanAstTransDetail astTransDetailIn = LoanAstTransDetail.createTransDtl(
                    loanAstIn, instalTransDetail, EventCodeEnum.TRANS_IN);
            // 机构资产转入处理本金
            astTransDetailIn.transIn(transPrinAmt, cost);

            if (loanAstIn.isNew()) {
                // 对于新创建的机构资产，初始化机构收益
                LoanAstProfitTransDetail
                        .createTransDtl(loanAstIn.getLoanAstProfit(loanAstIn.getTermNo(), ProfitTypeEnum.OWN),
                                instalTransDetail, EventCodeEnum.TRANS_IN);
            }

            // 处理转出
            LoanAstTransDetail astTransDetailOut = LoanAstTransDetail.createTransDtl(loanAst, instalTransDetail,
                    EventCodeEnum.TRANS_OUT);
            // 机构资产转出处理本金
            astTransDetailOut.transferOut();
            // 分期交易处理转让本金
            instalTransDetail.instalTransfer(cost, transPrinAmt);
        }

        if (!isTransfer) {
            LOGGER.warn(MessageUtil.formatMsg("【{0}】对分期处理转入过程中，没有资产被转让", trTask));
            return false;
        }

        /**
         * 转让清算登记
         */
        planClearingManager.processClearingForLoanAst(
                loanTransDetail.getLoanInstalTransDetails(), TransCodeEnum.BUY_AST, transInTemplate.getAccountDate(),
                instalTransDetail -> {
                    Map<PlanClearingLogKey, Money> ret = Maps.newHashMap();
                    for (LoanAstTransDetail astTransDetail : instalTransDetail.getLoanAstTransDetails()) {
                        LoanAstLog astLog = astTransDetail.getActiveLog();
                        if (EventCodeEnum.TRANS_OUT == astLog.getEventCode()) {
                            continue;
                        }
                        PlanClearingLogKey key = new PlanClearingLogKey(astLog.getOwnOrg(), trTask.getTfrOut(),
                                astLog.getEventCode().getSceneCode());
                        if (!ret.containsKey(key)) {
                            ret.put(key, new Money());
                        }

                        ret.get(key).addTo(astLog.getClrgAmt());
                    }
                    return ret;
                });

        /**
         * 转让后，借据状态校验
         */
        boolean isAllTfr = true;
        for (LoanInstalDomain loanInstal : loan.getLoanInstalList()) {
            // 如果不是逾期或正常的分期，不做转让（对应过滤的状态是：结清，冲正，取消）
            if (!loanInstal.getStatus().isActive()) {
                continue;
            }

            LoanAstGroup instalAstGroup = loanInstal.filterLoanAst(
                    loanInstal.getAstOrg().getAstOwnership().getLenderOrgCode());
            if (CollectionUtils.isNotEmpty(instalAstGroup.getLoanAsts())) {
                isAllTfr = false;
                break;
            }
        }

        if (isAllTfr) {
            loan.getAstOrg().getAstOwnership().setOwnStatus(OwnStatusEnum.TRANS_OUT);
            loan.setCapitalNo(CapitalAstMatchEnum.DEFAULT.getCode());
            loan.setLayerName(null);
        }

        loanTransDetail.processBeforeUpdate();
        loanAstAssembler.store(loan, trTask.getTfrDt(), false);
        // 对于转入机构的第一次备份
        List<LoanBak> loanBaks = LoanTransDetail.buildBaks(loan, transInTemplate.getAccountDate());
        loanRepository.backup(loanBaks);
        LOGGER.info("[LoanAstTrManager]转让完成：" + trTask);

        return true;
    }

}
