/**
 * abssqr.com Inc.
 * Copyright (c) 2017-2019 All Rights Reserved.
 */
package com.abssqr.plat.core.service.astscreen;

import cn.hutool.core.util.IdcardUtil;
import com.abssqr.plat.common.dal.mysql.auto.resultmap.AstPoolItem;
import com.abssqr.plat.common.facade.enums.AstRiskClassEnum;
import com.abssqr.plat.common.facade.enums.CalcIntrTypeEnum;
import com.abssqr.plat.common.facade.enums.CertTypeEnum;
import com.abssqr.plat.common.facade.enums.RateTypeEnum;
import com.abssqr.plat.common.facade.enums.ScreenTypeEnum;
import com.abssqr.plat.common.facade.model.rule.AstScreenRule;
import com.abssqr.plat.common.model.domain.ast.AstConditions;
import com.abssqr.plat.common.model.domain.ast.Pack;
import com.abssqr.plat.common.model.utils.TextUtil;
import com.abssqr.plat.common.util.loggers.CommonLoggers;
import com.general.system.common.model.Money;
import com.general.system.common.util.DateTimeUtil;
import com.general.system.common.util.DateTimeUtil.BetweenType;
import com.general.system.common.util.LogUtil;
import com.general.system.common.util.SystemDateUtil;
import com.general.system.common.util.VarChecker;
import com.general.system.tool.util.ToolUtil;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang.builder.ToStringBuilder;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.springframework.stereotype.Component;

import java.math.BigDecimal;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import static org.apache.commons.lang.builder.ToStringStyle.SHORT_PREFIX_STYLE;

/**
 * 筛选规则匹配
 *
 * @author bangis.wangdf
 * @version com.abssqr.plat.biz.shared.handler: AstScreenRuleHandler.java, v 0.1 2019-06-16 21:13 bangis.wangdf Exp $
 */
@Component
public class AstScreenRuleChecker {

    private final static Logger LOGGER = CommonLoggers.AST_MATCH;


    /**
     * 筛选规则匹配
     *
     * @param acctDate
     * @param astPoolItem
     * @param pack
     * @param screenTypeEnum
     * @return
     */
    public boolean matchWithTransferPack(Date acctDate, AstPoolItem astPoolItem, Pack pack, ScreenTypeEnum screenTypeEnum) {
        Map<String, String> context = TextUtil.convertTxt2Map(astPoolItem.getContext());
        // 当前日期在 转让开始日与结束日之间
        boolean trCheck = DateTimeUtil.between(DateTimeUtil.getStandardDate(acctDate), pack.getTransferRule().getStartTrDate(),
                pack.getTransferRule().getStopTrDate(),
                BetweenType.se);
        if(trCheck) {
            Date minEndDate = pack.getMinEndDate() == null ? DateTimeUtil.addYears(SystemDateUtil.getSystemDate(), -1)
                : pack.getMinEndDate();
            VarChecker.checkNotNull(pack.getMaxEndDate(), "[PlanAstMatchJob] pack.MaxEndDate不能为空，防止资产出头[{0}]", pack);
            if (ToolUtil.isNotEmpty(astPoolItem.getEndDate())
                && DateTimeUtil.between(astPoolItem.getEndDate(), minEndDate, pack.getMaxEndDate(), BetweenType.se)) {
                if (this.astScreenRuleMatch(acctDate, astPoolItem, context, pack,
                        pack.getTransferRule().getAstScreenRuleList(),screenTypeEnum)) {
                    return true;
                }
            }
        }
        LogUtil.info(LOGGER, "[AstScreenRuleChecker.match=false]-[{0}]-[{1}]",
            ToStringBuilder.reflectionToString(astPoolItem, SHORT_PREFIX_STYLE), pack);
        return false;
    }
    
    /**
     * 筛选规则匹配
     * @param acctDate
     * @param astPoolItem
     * @param context
     * @param pack
     * @param screenRules
     * @param screenTypeEnum
     * @return
     */
    public boolean astScreenRuleMatch(Date acctDate, AstPoolItem astPoolItem, Map<String, String> context,
                                 Pack pack, List<AstScreenRule> screenRules, ScreenTypeEnum screenTypeEnum) {
        List<AstScreenRule> astScreenRuleList = screenRules.stream().filter(
                item->item.getScreenType() == screenTypeEnum).collect(Collectors.toList());
        if (ToolUtil.isEmpty(astScreenRuleList)){
            return true;
        }
        for (AstScreenRule astScreenRule : astScreenRuleList) {
            //产品
            boolean prod = StringUtils.equals(astScreenRule.getAbssqrProdCode(), astPoolItem.getProdNo());

            // 最大年龄.
            boolean age = true;
            if (astScreenRule.getMinAge() != null || astScreenRule.getMaxAge() != null) {
                if (CertTypeEnum.getByCode(context.get(AstConditions.CERT_TYPE)) == CertTypeEnum.CTE_0) {
                    if (StringUtils.isNotBlank(context.get(AstConditions.CERT_NO)) && !StringUtils.equals(
                        context.get(AstConditions.CERT_NO), "null")) {
                        int certAge = IdcardUtil.getAgeByIdCard(context.get(AstConditions.CERT_NO), acctDate);
                        age = match(certAge, astScreenRule.getMinAge(), astScreenRule.getMaxAge());
                    } else {
                        age = match(null, astScreenRule.getMinAge(), astScreenRule.getMaxAge());
                    }
                }
            }

            // 到期日.
            boolean enDate = match(astPoolItem.getEndDate(), astScreenRule.getEarliestEndDate(),
                astScreenRule.getLatestEndDate());
            // 原始权益机构
            boolean lenderOrgCode = StringUtils.equals(astPoolItem.getLenderOrgCode(),
                astScreenRule.getLenderOrgCode());

            //  期次.
            boolean termCnt = match(astPoolItem.getTermCnt(), astScreenRule.getMinTerm(), astScreenRule.getMaxTerm());

            //  剩余天数.
            boolean remainDay = match(astPoolItem.getRemainDays(), astScreenRule.getMinRemainDays(),
                astScreenRule.getMaxRemainDays());

            // 利率.
            VarChecker.checkNotNull(context.get(AstConditions.INTER_RATE), "待筛选资产利率不能为null:[{0}]", astPoolItem.getAstNo());
            VarChecker.checkNotNull(context.get(AstConditions.RATE_TYPE), "待筛选资产利率类型不能为null:[{0}]", astPoolItem.getAstNo());
            BigDecimal rate = new BigDecimal(context.get(AstConditions.INTER_RATE));
            RateTypeEnum rateType = RateTypeEnum.getByCode(context.get(AstConditions.RATE_TYPE));

            rate = calcRate(rate, rateType, pack.getCalcIntrType());
            boolean interRate = match(rate, astScreenRule.getSingleMinInterRate(),
                astScreenRule.getSingleMaxInterRate());

            // 金额.
            boolean amt = match(astPoolItem.getPrinBal(), astScreenRule.getSingleMinAmount(),
                astScreenRule.getSingleMaxAmount());
            ///**
            // * region 地区.
            // */
            //private List<String> regionList;
            //  资产状态.

            boolean astClzStatus = true;
            if (CollectionUtils.isNotEmpty(astScreenRule.getAssetStatusList())) {
                astClzStatus = astScreenRule.getAssetStatusList().contains(
                    AstRiskClassEnum.getByCode(astPoolItem.getRiskCl()));
            }
            if (prod && age && enDate && lenderOrgCode && termCnt && remainDay && interRate && amt
                && astClzStatus) {
                return true;
            } else {
                LogUtil.info(LOGGER, "[AstScreenRuleChecker.match=false]-[prod:{0},age:{1},enDate:{2},lenderOrgCode:{3},termCnt:{4},remainDay:{5},interRate:{6},amt:{7},astClzStatus:{8},AstNo:{9}.{10}-{11}-{12}]", prod,
                    age, enDate, lenderOrgCode, termCnt, remainDay, interRate, amt
                    , astClzStatus,astPoolItem.getAstNo(),astPoolItem.getTermNo(),astScreenRule.getScreenRuleNo(),pack.getPackCode());
            }
        }
        return false;
    }

    private BigDecimal calcRate(BigDecimal rate, RateTypeEnum rateType, CalcIntrTypeEnum calcIntrType) {
        VarChecker.checkNotNull(rate, "astPool利率未指定");
        VarChecker.checkNotNull(rateType, "astPool利率类型未指定");
        BigDecimal result = null;
        switch (rateType) {
            case RT01:
                result = rate;
                break;
            case RT02:
                BigDecimal _12 = new BigDecimal(12);
                result = rate.multiply(_12);
                break;
            case RT03:
                VarChecker.checkNotNull(calcIntrType, "Pact.calcIntrType未指定");
                result = rate.multiply(calcIntrType.getAnnualDays());
                break;
            default:
                VarChecker.checkNotNull(null, "不支持的利率类型={0}", rateType);
                break;
        }
        return result;
    }

    /**
     * 金额
     *
     * @param gal
     * @param min
     * @param max
     * @return
     */
    private boolean match(Money gal, Money min, Money max) {
        if (min == null && max == null) {
            return true;
        }
        if (gal == null) {
            return false;
        }
        boolean resultMax = true;
        boolean resultMin = true;
        if (max != null) {
            resultMax = gal.lesserOrEquals(max);
        }
        if (min != null) {
            resultMin = gal.greaterOrEquals(min);
        }
        return resultMin && resultMax;
    }

    /**
     * 利率
     *
     * @param gal
     * @param min
     * @param max
     * @return
     */
    private boolean match(BigDecimal gal, BigDecimal min, BigDecimal max) {
        if (min == null && max == null) {
            return true;
        }
        if (gal == null) {
            return false;
        }
        boolean resultMax = true;
        boolean resultMin = true;
        if (max != null) {
            resultMax = gal.compareTo(max) <= 0;
        }
        if (min != null) {
            resultMin = gal.compareTo(min) >= 0;
        }
        return resultMin && resultMax;
    }

    /**
     * 到期日.
     *
     * @param gal
     * @param min
     * @param max
     * @return
     */
    private boolean match(Date gal, Date min, Date max) {
        if (min == null && max == null) {
            return true;
        }
        if (gal == null) {
            return false;
        }
        boolean resultMax = true;
        boolean resultMin = true;
        if (max != null) {
            resultMax = DateTimeUtil.calcDayDiff(max,gal)>=0;
        }
        if (min != null) {
            resultMin = DateTimeUtil.calcDayDiff(gal,min ) >=0;
        }
        return resultMin && resultMax;
    }

    /**
     * 期次、 剩余天数
     *
     * @param gal
     * @param min
     * @param max
     * @return
     */
    private boolean match(Integer gal, Integer min, Integer max) {
        if (min == null && max == null) {
            return true;
        }
        if (gal == null) {
            return false;
        }
        boolean resultMax = true;
        boolean resultMin = true;
        if (max != null) {
            resultMax = gal <= max;
        }
        if (min != null) {
            resultMin = gal >= min;
        }
        return resultMin && resultMax;
    }

}
