/**
 * abssqr.com Inc.
 * Copyright (c) 2017-2019 All Rights Reserved.
 */
package com.abssqr.plat.core.service.plan;

import com.abssqr.plat.common.facade.constants.AbssqrConstants;
import com.abssqr.plat.common.facade.enums.CalcIntrTypeEnum;
import com.abssqr.plat.common.facade.enums.CalcTypeEnum;
import com.abssqr.plat.common.facade.enums.EventCodeEnum;
import com.abssqr.plat.common.facade.enums.FeeScaleTypeEnum;
import com.abssqr.plat.common.facade.enums.TitleTypeEnum;
import com.abssqr.plat.common.facade.model.plan.PlanAcctTitle;
import com.abssqr.plat.common.facade.model.plan.PlanBase;
import com.abssqr.plat.common.facade.utils.CalcUtil;
import com.abssqr.plat.common.model.domain.loanTrans.TransInfoTemplate;
import com.abssqr.plat.common.model.domain.plan.PlanAcctTitleTransDetail;
import com.abssqr.plat.common.model.repo.loanTrans.LoanInstalLogRepository;
import com.abssqr.plat.core.service.system.TimeStampUtil;
import com.general.system.common.model.Money;
import com.general.system.common.util.DateTimeUtil;
import com.general.system.common.util.LogUtil;
import com.general.system.common.util.MessageUtil;
import com.general.system.common.util.SystemDateUtil;
import com.general.system.common.util.VarChecker;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.math.BigDecimal;
import java.util.Date;
import java.util.List;
import java.util.Map;

/**
 * 计划结息费
 *
 * @author zhenxuan.luo
 * @version com.abssqr.plat.core.service.plan: PlanSettleManager.java, v 0.1 2019-06-04 01:28 zhenxuan.luo Exp $
 */
@Component
public class PlanSettleManager {
    
    private final static Logger LOGGER = LoggerFactory.getLogger(PlanSettleManager.class);
    
    @Autowired
    private LoanInstalLogRepository loanInstalLogRepository;
    
    public static PlanAcctTitleTransDetail buildPlanAcctTitleTransDetail(Date accountDate, PlanAcctTitle planAcctTitle,
                                                                         boolean isReal) {
        TransInfoTemplate transInfoTemplate = new TransInfoTemplate();
        transInfoTemplate.setTransTime(SystemDateUtil.getSystemDate());
        transInfoTemplate.setAccountDate(DateTimeUtil.getStandardDate(accountDate));
        transInfoTemplate.setEventCode(EventCodeEnum.TITLE_SETTLE);
        String bizNo = genBizNo(planAcctTitle, accountDate);
        transInfoTemplate.setOutBizNo(bizNo);
        transInfoTemplate.setBizNo(bizNo);
        transInfoTemplate.setTimestamp(TimeStampUtil.getDefaultTimeStamp());
        
        return PlanAcctTitleTransDetail
                       .createTxnDtl(planAcctTitle, transInfoTemplate,
                               isReal);
    }
    
    private static String genBizNo(PlanAcctTitle planAcctTitle, Date accountDate) {
        return MessageUtil.formatMsg("{0}-{1}-{2}", DateTimeUtil.formatYMD(accountDate), planAcctTitle.getPlanNo(),
                planAcctTitle.getTitleNo());
    }
    
    /**
     * 计算基数金额
     *
     * @param planAcctTitleList
     * @return
     */
    public RefInvestAmt calcRefInvestment(List<PlanAcctTitle> planAcctTitleList, Date accountDate) {
        RefInvestAmt refInvestAmt = new RefInvestAmt();
        
        // 初始化计息费的基数
        for (PlanAcctTitle planAcctTitle : planAcctTitleList) {
            
            if (null != accountDate) {
                /**
                 * 不计息费的场景
                 * 1. 未到生效日
                 * 2. 生效日下一日
                 */
                Date settleDay = getSettleDay(planAcctTitle.getEffectDate(), accountDate);
                if (settleDay == null) {
                    continue;
                } else if (DateTimeUtil.calcDayDiff(accountDate, planAcctTitle.getEffectDate()) != 0
                           && DateTimeUtil.calcDayDiff(settleDay, planAcctTitle.getEffectDate()) == 0) {
                    /**
                     * 生效日=会计日：默认计息费
                     * 除此之外，如果settleDay=生效日，说明是生效日的下一天，不结息
                     */
                    continue;
                }
            }
            
            if (planAcctTitle.getTitleType().isInvest()) {
                refInvestAmt.investBal.put(planAcctTitle.getRefNo(), planAcctTitle.getBal().clone());
            }
            
            switch (planAcctTitle.getTitleType()) {
                case PRIOR_INVESTED:
                    refInvestAmt.priorInitIvstAmt.addTo(planAcctTitle.getInitAmt());
                    refInvestAmt.priorIvstBal.addTo(planAcctTitle.getBal());
                    // no break
                case SECND_INVESTED:
                    refInvestAmt.allInitIvstAmt.addTo(planAcctTitle.getInitAmt());
                    refInvestAmt.allIvstBal.addTo(planAcctTitle.getBal());
                    break;
                default:
                    break;
            }
        }
        
        return refInvestAmt;
    }
    
    /**
     * 对指定会计日，进行结息费
     *
     * @param accountDate
     * @param planAcctTitleList
     */
    public List<PlanAcctTitleTransDetail> settle(PlanBase planEntity, Date accountDate,
                                                 List<PlanAcctTitle> planAcctTitleList, boolean isReal) {
        
        accountDate = DateTimeUtil.getStandardDate(accountDate);
        
        List<PlanAcctTitleTransDetail> ret = Lists.newArrayList();
        for (PlanAcctTitle planAcctTitle : planAcctTitleList) {
            List<Date> settleAcctountDates = Lists.newArrayList();
            if (isReal) {
                // 补计息，为了拿到每天的计息流水
                Date settleDay = getSettleDay(planAcctTitle.getEffectDate(), accountDate);
                if (null == settleDay) {
                    LogUtil.info(LOGGER, "{0}[{1}]的科目[{2}]在[{3}]不需要结息:未到生效日[{4}]，退出",
                            planEntity.getPlanType().getDesc(), planEntity.getPlanNo(),
                            planAcctTitle.getTitleNo(), accountDate, planEntity.getPlanType());
                    continue;
                }
                Date beginDay = getSettleBeginDay(settleDay, planAcctTitle);
                if (null == beginDay) {
                    LogUtil.info(LOGGER, "{0}[{1}]的科目[{2}]在[{3}]不需要结息:已结息[{4}]-[{5}]，退出",
                            planEntity.getPlanType().getDesc(), planEntity.getPlanNo(),
                            planAcctTitle.getTitleNo(), accountDate, settleDay, planAcctTitle.getLastSettleDate());
                    continue;
                }
                while (DateTimeUtil.calcDayDiff(accountDate, beginDay) >= 0) {
                    settleAcctountDates.add(beginDay);
                    beginDay = DateTimeUtil.addDays(beginDay, 1);
                }
            } else {
                settleAcctountDates.add(accountDate);
            }
            
            settleAcctountDates.sort(Date::compareTo);
            
            for (Date settleAccountDate : settleAcctountDates) {
                RefInvestAmt refInvestAmt = this.calcRefInvestment(planAcctTitleList, settleAccountDate);
                PlanAcctTitleTransDetail transDetail = settlePlanAcct(planEntity, settleAccountDate, refInvestAmt,
                        planAcctTitle, isReal);
                if (transDetail != null) {
                    transDetail.processBeforeUpdate();
                    ret.add(transDetail);
                }
            }
        }
        
        return ret;
    }
    
    /**
     * 计算计划科目结息费
     *
     * @param planEntity
     * @param accountDate
     * @param refInvestAmt
     * @param planAcctTitle
     * @param isReal
     * @return
     */
    private PlanAcctTitleTransDetail settlePlanAcct(PlanBase planEntity, Date accountDate,
                                                    RefInvestAmt refInvestAmt,
                                                    PlanAcctTitle planAcctTitle, boolean isReal) {
        
        PlanAcctTitleTransDetail planAcctTitleTransDetail = buildPlanAcctTitleTransDetail(accountDate, planAcctTitle,
                isReal);
        
        // 本金不用计
        if (planAcctTitle.getTitleType() == TitleTypeEnum.PRIOR_INVESTED
            || planAcctTitle.getTitleType() == TitleTypeEnum.SECND_INVESTED) {
            return planAcctTitleTransDetail;
        }
        // 固定金额或固定费率：不做计算，在科目初始化的时候已经算过了
        if (planAcctTitle.getCalcType() != CalcTypeEnum.ANNU_RATE) {
            return planAcctTitleTransDetail;
        }
        // 计息方式
        CalcIntrTypeEnum calcIntrType = planEntity.getCalcIntrType();
        
        // 计算计息天数
        // 先计算需要结息到哪天（含）
        Date settleDay = this.getSettleDay(planAcctTitle.getEffectDate(), accountDate);
        if (settleDay == null) {
            LogUtil.info(LOGGER, "{0}[{1}]的科目[{2}]在[{3}]不需要结息:未到生效日[{4}]，退出",
                    planEntity.getPlanType().getDesc(), planEntity.getPlanNo(),
                    planAcctTitle.getTitleNo(), accountDate, planEntity.getEffectDate());
            return null;
        }
        // 然后计算从哪天结息（含）
        Date beginDay = this.getSettleBeginDay(settleDay, planAcctTitle);
        if (beginDay == null) {
            LogUtil.info(LOGGER, "{0}[{1}]的科目[{2}]在[{3}]不需要结息:已结息[{4}]-[{5}]，退出",
                    planEntity.getPlanType().getDesc(), planEntity.getPlanNo(),
                    planAcctTitle.getTitleNo(), accountDate, settleDay, planAcctTitle.getLastSettleDate());
            return null;
        }
        
        long days = DateTimeUtil.calcDayDiff(settleDay, beginDay) + 1;
        VarChecker.checkArgument(days > 0, "{0}[{1}]的科目[{2}]在[{3}]结息异常，计息天数[{4}]必须大于0",
                planEntity.getPlanType().getDesc(), planEntity.getPlanNo(), planAcctTitle.getTitleNo(), accountDate,
                days);
        Money refAmt;
        if (planAcctTitle.getScaleType() == null) {
            refAmt = refInvestAmt.getByScaleType(planAcctTitle.getRefNo());
        } else {
            refAmt = refInvestAmt.getByScaleType(planAcctTitle.getScaleType());
        }
        
        // 先乘再除  科目利率 *  * (结息日 - 起息日)
        BigDecimal yearRateAmt = CalcUtil.getRealRate(planAcctTitle.getRate()).multiply(refAmt.getAmount())
                                         .multiply(new BigDecimal(days));
        
        // 乘完之后再做日化处理，减少精度损失
        BigDecimal addTotleAmt = yearRateAmt.divide(calcIntrType.getAnnualDays(), AbssqrConstants.DECIMAL_SCALE,
                AbssqrConstants.DECIMAL_ROUNDING_MODE);
        planAcctTitleTransDetail.setSettleDate(settleDay);
        planAcctTitleTransDetail.entry(addTotleAmt);
        return planAcctTitleTransDetail;
    }
    
    /**
     * 从哪天开始结息
     *
     * @param settleDay
     * @return
     */
    private Date getSettleBeginDay(Date settleDay, PlanAcctTitle planAcctTitle) {
        // 当天已经结过息
        if (planAcctTitle.getLastSettleDate() != null
            && DateTimeUtil.calcDayDiff(planAcctTitle.getLastSettleDate(), settleDay) >= 0) {
            return null;
        }
        Date beginDay;
        if (null == planAcctTitle.getLastSettleDate()
            || DateTimeUtil.calcDayDiff(planAcctTitle.getLastSettleDate(), planAcctTitle.getEffectDate()) < 0) {
            beginDay = planAcctTitle.getEffectDate();
        } else {
            // 取最后一次结息日+1
            beginDay = DateTimeUtil.addDays(planAcctTitle.getLastSettleDate(), 1);
        }
        
        return beginDay;
    }
    
    /**
     * 计算当前计划结息到哪一天
     *
     * @param effectDate  起息日
     * @param accountDate
     * @return
     */
    public Date getSettleDay(Date effectDate, Date accountDate) {
        // 未到计划生效日，不结息
        if (DateTimeUtil.calcDayDiff(effectDate, accountDate) > 0) {
            return null;
        }
        // beginDay始终大于等于计划的生效日
        // 算头不算尾
        if (DateTimeUtil.calcDayDiff(accountDate, effectDate) == 0) {
            // 生效日当天结息
            return accountDate;
        } else {
            // 生效日后结上一天利息（第二天不结息）
            return DateTimeUtil.addDays(accountDate, -1);
        }
    }
    
    /**
     * 基数金额
     */
    public class RefInvestAmt {
        //优先级初始资本
        private Money priorInitIvstAmt = new Money(0);
        //优先级资本余额
        private Money priorIvstBal     = new Money(0);
        //优先加劣后初始资本
        private Money allInitIvstAmt   = new Money(0);
        //优先加劣后资本余额
        private Money allIvstBal       = new Money(0);
        
        //投入资本余额
        private Map<String, Money> investBal = Maps.newHashMap();
        
        /**
         * 根据基数类型获取基数
         *
         * @param scaleType
         * @return
         */
        public Money getByScaleType(FeeScaleTypeEnum scaleType) {
            switch (scaleType) {
                case START:
                    return allInitIvstAmt;
                case REMAIN:
                    return allIvstBal;
                case PRISTART:
                    return priorInitIvstAmt;
                case PRIREMAIN:
                    return priorIvstBal;
                default:
                    return new Money(0);
            }
        }
        
        public Money getByScaleType(String refNo) {
            Money refAmt = investBal.get(refNo);
            if (refAmt == null) {
                return new Money(0);
            }
            return refAmt;
        }
    }
    
}

