/**
 * abssqr.com Inc. 
 * Copyright (c) 2017-2019 All Rights Reserved. 
 */
package com.abssqr.plat.core.service.report;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.date.DateUtil;
import cn.hutool.core.util.StrUtil;
import com.abssqr.plat.common.facade.enums.EventCodeEnum;
import com.abssqr.plat.common.facade.enums.TitleTypeEnum;
import com.abssqr.plat.common.facade.model.plan.PlanBase;
import com.abssqr.plat.common.facade.model.report.PlanReportItem;
import com.abssqr.plat.common.facade.model.report.ReportField;
import com.abssqr.plat.common.facade.model.report.wk.WkAstReportFields;
import com.abssqr.plat.common.facade.utils.CalcUtil;
import com.abssqr.plat.common.model.domain.plan.PlanAcctTitleLog;
import com.abssqr.plat.common.model.repo.plan.PlanAcctTitleLogRepo;
import com.abssqr.plat.common.model.repo.plan.PlanRepo;
import com.abssqr.plat.common.model.repo.report.PlanReportRepo;
import com.abssqr.plat.core.service.plan.PlanManager;
import com.abssqr.plat.core.service.util.AnalyseReportUtil;
import com.abssqr.plat.core.service.util.RptBuilderUtil;
import com.general.system.common.model.Money;
import com.general.system.common.util.VarChecker;
import com.general.system.tool.util.ToolUtil;
import com.google.common.collect.Lists;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.math.BigDecimal;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

/**
 * 为统计页面组装相关报表数据
 * @author hanfei
 * @version $Id: AnalyseSpvReportManager.java, v 0.1 2019-09-27 10:56 AM hanfei Exp $
 */
@Component
public class AnalyseSpvReportManager {
    
    @Autowired
    private PlanReportRepo       planReportRepo;
    @Autowired
    private PlanManager          planManager;
    @Autowired
    private PlanRepo             planRepo;
    @Autowired
    private PlanAcctTitleLogRepo planAcctTitleLogRepo;
    /**
     * 查询计划
     * @param planNo
     * @return
     */
    public PlanBase checkGetPlan(String planNo) {
        PlanBase planBase = planRepo.getPlanByNo(planNo);
        VarChecker.checkNotNull(planBase, "找不到对应项目 [{0}]", planNo);
        return planBase;
    }
    /**
     * 查询完整计划
     * @param planNo
     * @return
     */
    public PlanBase checkGetFullPlan(String planNo) {
        PlanBase planBase = planManager.getPlanFullByNo(planNo);
        VarChecker.checkNotNull(planBase, "找不到对应项目 [{0}]", planNo);
        return planBase;
    }
    /**
     * 查询计划对应的受让池计划
     * @param planNo
     * @return
     */
    public List<PlanBase> getPlanBySpv(String planNo){
        Map<String, List<PlanBase>> planMap = planManager.getPlanBySpvs(Lists.newArrayList(planNo));
        return planMap.get(planNo);
    }
    
    /**
     * 根据会计日 获取放款池 汇总统计承接池
     * @param accountDate 会计日
     * @param spvPlanNo  放款池编号
     * @param fields    统计字段
     * @param havePlan 是否包含对应的承接池
     * @return
     */
    public Map<String,PlanReportItem> querySpvStatPlanDayItem(Date accountDate, String spvPlanNo,
                                                          List<ReportField> fields, boolean havePlan){
        // 查询放款池计划统计信息
        Map<String, PlanReportItem> spvReportItemMap =
                planReportRepo.queryRptItemsByPlanNoAndAccDate(accountDate, spvPlanNo, fields);
        if (!havePlan || ToolUtil.isEmpty(spvReportItemMap)){
            return spvReportItemMap;
        }
        List<PlanBase> plans = this.getPlanBySpv(spvPlanNo);
        // 查询对应受让池计划信息
        if (ToolUtil.isEmpty(plans)){
            return spvReportItemMap;
        }
        // 汇总数据
        for(PlanBase plan : plans) {
            Map<String, PlanReportItem>  reportItemMap =
                    planReportRepo.queryRptItemsByPlanNoAndAccDate(accountDate, plan.getPlanNo(), fields);
            if (ToolUtil.isEmpty(reportItemMap)){
                continue;
            }
            for(ReportField field : fields){
                // 合并统计项
                AnalyseReportUtil.summaryByField(plan.getPlanNo(), spvReportItemMap, reportItemMap, field);
            }
        }
        return spvReportItemMap;
    }
    
    /**
     * 获取 计划维度区间合计 汇总统计受让池
     * @param startDate
     * @param endDate
     * @param spvPlanNo  放款池编号
     * @param fields    统计字段
     * @param havePlan 是否包含对应的承接池
     * @return
     */
    public Map<String,PlanReportItem> querySpvStatPlanSectionItem(Date startDate, Date endDate, String spvPlanNo,
                                                                  List<ReportField> fields, boolean havePlan){
        return this.querySpvStatPlanSectionItem(startDate, endDate, spvPlanNo, fields, havePlan, null);
    }
    /**
     * 获取 计划维度区间合计 汇总统计受让池 重置spare时间长度
     * @param startDate
     * @param endDate
     * @param spvPlanNo  放款池编号
     * @param fields    统计字段
     * @param havePlan 是否包含对应的承接池
     * @param spare     统计区间长度 传入值会根据实际统计时间段计算
     * @return
     */
    public Map<String,PlanReportItem> querySpvStatPlanSectionItem(Date startDate, Date endDate, String spvPlanNo,
                                                                  List<ReportField> fields, boolean havePlan, AtomicInteger spare){
        
        // 根据时间区间获取数据中的开始时间并重新赋值
        Date dateStartDate = planReportRepo.queryReportMinDateByPlanNoSection(startDate, endDate, spvPlanNo, fields);
        if (ToolUtil.isNotEmpty(dateStartDate) && ToolUtil.isNotEmpty(spare)){
            // 设置区间时长
            spare.set(Long.valueOf(DateUtil.betweenDay(dateStartDate, endDate, true) + 1 ).intValue());
        }
        // 查询放款池计划统计信息
        Map<String, PlanReportItem> spvReportItemMap =
                planReportRepo.queryRptItemsByPlanNoSection(startDate, endDate, spvPlanNo, fields);
        if (!havePlan || ToolUtil.isEmpty(spvReportItemMap)){
            return spvReportItemMap;
        }
        List<PlanBase> plans = this.getPlanBySpv(spvPlanNo);
        // 查询对应受让池计划信息
        if (ToolUtil.isEmpty(plans)){
            return spvReportItemMap;
        }
        // 汇总数据
        for(PlanBase plan : plans) {
            Map<String, PlanReportItem>  reportItemMap =
                    planReportRepo.queryRptItemsByPlanNoSection(startDate, endDate, plan.getPlanNo(), fields);
            if (ToolUtil.isEmpty(reportItemMap)){
                continue;
            }
            for(ReportField field : fields){
                // 合并统计项
                AnalyseReportUtil.summaryByField(plan.getPlanNo(), spvReportItemMap, reportItemMap, field);
            }
        }
        return spvReportItemMap;
    }
    
    /**
     * 计算回款率（回收率）
     * 放款池、承接池近30天实际偿还额和/近30日每日资产剩余本金和
     * @param startDate
     * @param endDate
     * @param planNo
     * @return
     */
    public BigDecimal getRepayRate(Date startDate, Date endDate, String planNo) {
        List<ReportField> amtFields = Lists.newArrayList(
                WkAstReportFields.WK039,
                WkAstReportFields.WK040,
                WkAstReportFields.WK041,
                WkAstReportFields.WK017
        );
        
        // 查询放款池计划及受让池计划的对应统计项数据汇总
        Map<String, PlanReportItem> itemRepayMap = this.querySpvStatPlanSectionItem(
                startDate, endDate, planNo, amtFields, true);
        List<ReportField> repayFields = Lists.newArrayList(
                WkAstReportFields.WK039,
                WkAstReportFields.WK040,
                WkAstReportFields.WK041
        );
        /**
         * 计算还款率  还款金额综合 / 总贷款本金余额
         */
        BigDecimal denominator = AnalyseReportUtil.getItemBig(itemRepayMap, WkAstReportFields.WK017);
        return CalcUtil.dividePercRate(
                AnalyseReportUtil.getItemBig(itemRepayMap, repayFields), denominator);
    }
    
    /**
     * 早偿率
     * （放款池、承接池）近30日早偿金额和/近30日每日放款池、承接吃资产剩余本金和
     * @param startDate
     * @param endDate
     * @param planNo
     * @return
     */
    public BigDecimal getPreRepayRate(Date startDate, Date endDate, String planNo) {
        List<ReportField> amtFields = Lists.newArrayList(
                WkAstReportFields.WK042,
                WkAstReportFields.WK043,
                WkAstReportFields.WK017
        );
        // 查询放款池计划及受让池计划的对应统计项数据汇总
        Map<String, PlanReportItem> itemRepayMap = this.querySpvStatPlanSectionItem(
                startDate, endDate, planNo, amtFields, true);
        List<ReportField> repayFields = Lists.newArrayList(
                WkAstReportFields.WK042,
                WkAstReportFields.WK043
        );
        BigDecimal denominator = AnalyseReportUtil.getItemBig(itemRepayMap, WkAstReportFields.WK017);
        return CalcUtil.dividePercRate(
                AnalyseReportUtil.getItemBig(itemRepayMap, repayFields), denominator);
    }
    
    /**
     * 逾期率 按日期加权计算
     * （资产）放款池、承接池报表的逾期剩余本金和/ 日期区间 每日放款池、承接池资产剩余本金和
     * @param startDate
     * @param endDate
     * @param planNo
     * @return
     */
    public BigDecimal getOvdRate(Date startDate, Date endDate, String planNo) {
        List<ReportField> amtFields = Lists.newArrayList(
                WkAstReportFields.WK012,
                WkAstReportFields.WK017
        );
        // 查询放款池计划及受让池计划的对应统计项数据汇总
        Map<String, PlanReportItem> itemRepayMap = this.querySpvStatPlanSectionItem(
                startDate, endDate, planNo, amtFields, true);
        BigDecimal denominator = AnalyseReportUtil.getItemBig(itemRepayMap, WkAstReportFields.WK017);
        return CalcUtil.dividePercRate(
                AnalyseReportUtil.getItemBig(itemRepayMap, WkAstReportFields.WK012), denominator );
    }
    
    /**
     * 平均闲置率
     * 放款池、承接池 日可用资金总额/放款池+承接池信托规模  成立至今的加权平均
     * @param endDate
     * @param spvPlan
     * @return
     */
    public BigDecimal getAvgIdleRateRate(Date endDate, PlanBase spvPlan) {
        // 本金科目
        List<TitleTypeEnum> investedEnums =
                CollectionUtil.newArrayList(TitleTypeEnum.INVESTED, TitleTypeEnum.PRIOR_INVESTED, TitleTypeEnum.SECND_INVESTED);
        // 获取所有兑付流水
        List<PlanAcctTitleLog> spvTitleLogs = planAcctTitleLogRepo.getByPlanAndEvent(
                null, spvPlan.getPlanNo(), EventCodeEnum.TITLE_ALLOC, investedEnums);
        // 放款池存续规模累加合计
        Money planTotalAmt = RptBuilderUtil.getInvestedWeight(endDate, spvPlan, spvTitleLogs);
        
        // 查询对应受让池计划
        List<PlanBase> planBases = this.getPlanBySpv(spvPlan.getPlanNo());
        if (ToolUtil.isNotEmpty(planBases)){
            planBases.forEach(planBase -> {
                // 获取所有兑付流水
                List<PlanAcctTitleLog> planTitleLogs = planAcctTitleLogRepo.getByPlanAndEvent(
                        null, planBase.getPlanNo(), EventCodeEnum.TITLE_ALLOC, investedEnums);
        
                // 追加对应承接池计划的存续规模累加合计
                planTotalAmt.addTo(RptBuilderUtil.getInvestedWeight(endDate, planBase, planTitleLogs));
            });
        }
        // 查询闲置资金余额
        List<ReportField> amtFields = Lists.newArrayList(
                WkAstReportFields.WK001
        );
        // 闲置资金累加求和 查询放款池计划及受让池计划的对应统计项数据汇总
        Map<String, PlanReportItem> itemRepayMap = this.querySpvStatPlanSectionItem(
                spvPlan.getBeginDate(), endDate, spvPlan.getPlanNo(), amtFields, true);
        return CalcUtil.dividePercRate(
                AnalyseReportUtil.getItemBig(itemRepayMap, amtFields), planTotalAmt.getAmount());
    }

    /**
     * 获取 放款池计划 按日期维度获取统计数据
     * @param startDate
     * @param endDate
     * @param spvPlanNo
     * @param fields
     * @return
     */
    public Map<Date, List<PlanReportItem>> querySpvStatDaySectionItem(Date startDate, Date endDate, String spvPlanNo,
                                                              List<ReportField> fields, boolean havePlan){
        // 查询放款池计划 按日期汇总
        List<PlanReportItem> spvReportItems =
                planReportRepo.queryDayRptItemsByPlanNoSection(startDate, endDate, fields, CollectionUtil.newArrayList(spvPlanNo));
        if (ToolUtil.isEmpty(spvReportItems)){
            return null;
        }
        Map<Date, List<PlanReportItem>> spvDayItemMap =
                spvReportItems.stream().collect(Collectors.groupingBy(PlanReportItem::getAccountDate));
        if (!havePlan){
            return spvDayItemMap;
        }
        List<PlanBase> plans = this.getPlanBySpv(spvPlanNo);
        // 查询对应受让池计划信息
        if (ToolUtil.isEmpty(plans)){
            return spvDayItemMap;
        }
        // 对应受让池计划编号
        List<String> planNos = plans.stream().map(PlanBase::getPlanNo).collect(Collectors.toList());
    
        // 查询放款池对应的受让池计划 按日期汇总数据
        List<PlanReportItem> planReportItems =
                planReportRepo.queryDayRptItemsByPlanNoSection(startDate, endDate, fields, planNos);
        if (ToolUtil.isEmpty(planReportItems)){
            return spvDayItemMap;
        }
        // 按日期分组转Map
        Map<Date, List<PlanReportItem>> planDayItemMap =
                planReportItems.stream().collect(Collectors.groupingBy(PlanReportItem::getAccountDate));
        // 按日期汇总数据
        for(PlanReportItem spvDayItem : spvReportItems) {
            List<PlanReportItem> planDayItems = planDayItemMap.get(spvDayItem.getAccountDate());
            if (ToolUtil.isEmpty(planDayItems)){
                continue;
            }
            for(PlanReportItem planItem : planDayItems){
                // 相同统计项汇总
                if (StrUtil.equals(spvDayItem.getRptItemCode(), planItem.getRptItemCode())){
                    // 合并统计项
                    AnalyseReportUtil.summary(spvDayItem, planItem);
                }
            }
        }
        return spvDayItemMap;
    }
    
}
